#include <stdio.h>
#include <string.h>
#include <qdos.h>
#include <sys/socket.h>
#include <sys/time.h>

#include "qsockets.h"
#include <errno.h>

#define INC_SET 1
#define SET_SET 2

#ifndef max
# define max(x,y) ((x) < (y) ? (y) : (x))
#endif

int _fd_isset(int fd, fd_set *set)
{
    short i;
    int res = 0;

    if(fd < 256) 
    {
        fd = getchid(fd);
    }
    
    for(i = 0; i < FD_SETSIZE; i++)
    {
        if(set->fd[i] == fd)
        {
            res = (set->sts[i] & SET_SET);
            break;
        }
    }
    return res;
}

void _fd_clr(int fd, fd_set *set)
{
    short i;
    if(fd < 256) 
    {
        fd = getchid(fd);
    }
    
    for(i = 0; i < FD_SETSIZE; i++)
    {
        if(set->fd[i] == fd)
        {
            set->fd[i] = set->sts[i] = 0;
            break;
        }
    }

}

void _fd_set(int fd, fd_set *set)
{
    int f = -1;
    short i;
    if(fd < 256) 
    {
        fd = getchid(fd);
    }
    
    for(i = 0; i < FD_SETSIZE; i++)
    {
        if(f == -1 && set->fd[i] == 0)
        {
            f = i;
        }
        else if(set->fd[i] == fd)
        {
            f = i;
            break;
        }
    }

    if(f != -1)
    {
        set->fd[f] = fd;
        set->sts[f] = INC_SET;
    }
}

int  select (int n, fd_set *rfds, fd_set *wfds, fd_set *efds,
             struct timeval *tv)
{
    int wait, nr, nw, i, rv = 0;
    int qv;    
    if(n < 0)
    {
        rv = EINVAL;
    }
    else
    {
        rv = 0;
        if(tv == NULL)
        {
            wait = -1;
        }
        else
        {
            wait = tv->tv_sec*50 + tv->tv_usec * 50 / 1000000;
        }
        
        do
        {
       
            nr = nw = rv = 0;
            for(i = 0; i < n && rv == 0; i++)
            {
                if(rfds)
                {
                    if(rfds->fd[i] &&  (rfds->sts[i] & INC_SET))
                    {
                        if((qv = io_pend(rfds->fd[i], 0)) == 0)
                        {
                            rfds->sts[i] |= SET_SET;
                            nr++;
                        }
                        else if(qv != -1)
                        {
                            rv = EBADF;
                            break;
                        }
                    }
                }

                if(wfds && rv == 0)
                {
                    if(wfds->fd[i] && (wfds->sts[i] & INC_SET))
                    {
                        wfds->sts[i] |= SET_SET;
                        nw++;
                    }
                }
                    
            }
            if(nr || nw || rv) break;
            if(wait > 0) wait--;
            mt_susjb(-1, 1, NULL);
        } while (wait);
    }

    if(qv == -1)
    {
        rv = EAGAIN;
    }
    
    if(rv)
    {
        errno = rv;
        rv = -1;
    }
    else
    {
        errno = 0;
        rv = max(nw,nr);
    }

    if(wait > 0)
    {
        tv->tv_sec = wait / 50;
        tv->tv_usec = (wait % 50) * (1000000/50);
    }
    return rv;
}


