#include <errno.h>
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <sys/un.h>
#ifdef __QDOS__
# define strncasecmp strnicmp
#endif
#ifndef offsetof
# define offsetof(sname,fname) ((long)&((sname *)0)->fname)
#endif
#ifndef max
# define max(a,b) ((a) > (b) ? (a) : (b))
#endif
#define SERVPORT 2220

static void display_inet(void  *s, void *d)
{
    char *s1,*d1;

    if((s1 = strdup(inet_ntoa(((struct sockaddr_in *)s)->sin_addr))))
    {
	if((d1 = strdup(inet_ntoa(((struct sockaddr_in *)d)->sin_addr))))
	{
	    fprintf(stderr, "s = %s.%d <--> p = %s.%d (%x %x)\n", 
		    s1, htons(((struct sockaddr_in *)s)->sin_port),
		    d1, htons(((struct sockaddr_in *)d)->sin_port),
		    ((struct sockaddr_in *)s)->sin_addr.s_addr,
                    ((struct sockaddr_in *)d)->sin_addr.s_addr);
	    free(d1);
	}
	free(s1);
    }
}

static void display_unix(void *s, void *d)
{
    fprintf(stderr, "unix domain socket:  %s\n", 
            ((struct sockaddr_un *)s)->sun_path);
}

int  getsocket(struct sockaddr *iname, size_t *nlen, int port,
               char *host, char * fn)
{
    struct hostent *h;
    int sock,domain;
           
    if(fn == NULL)
    {
        if(host == NULL)
        {
            host = malloc(256);
            if(gethostname (host, 256))
            {
                perror ("gethostname");
                exit(0);
            }
        }

        if ((h = (struct hostent *) gethostbyname (host)) == NULL)
        {
            perror ("gethostbyname");
            exit(0);
        }

        ((struct sockaddr_in *)iname)->sin_family = h->h_addrtype;
        ((struct sockaddr_in *)iname)->sin_port = htons (port);
        ((struct sockaddr_in *)iname)->sin_addr =
            *((struct in_addr *) h->h_addr);
        *nlen = sizeof(struct sockaddr_in);
        domain = AF_INET;
    }
    else
    {
#ifndef QDOS
        unlink(fn);
#endif
        ((struct sockaddr_un *)iname)->sun_family = AF_UNIX;
        strcpy (((struct sockaddr_un *)iname)->sun_path, fn);
        *nlen = SUN_LEN((struct sockaddr_un *)iname);
        domain = AF_UNIX;      
    }

    if ((sock = socket (domain, SOCK_STREAM, 0)) == -1)
    {
	perror ("socket");
	exit (EXIT_FAILURE);
    }
    return sock;
}


int main (int argc, char **argv)
{
    int sock, asock;
    char msg [1024];
    struct sockaddr *iname, *oname;
    short done;
    int c;
    char *fn = NULL;
    int port = SERVPORT;
    int olen, nlen;
    void (*dfunc)(void *s, void *d);
    
    olen = nlen =
        max(sizeof(struct sockaddr_in), sizeof(struct sockaddr_un));    
    iname = malloc(nlen);
    oname = malloc(nlen);    

    while((c = getopt(argc, argv, "p:u:")) != EOF)
    {
        switch(c)
        {
            case 'p':
                port = atoi(optarg);
                break;
            case 'u':
                fn = optarg;
                break;
            default:
                fputs("ts [-p port] [-u name]\n", stderr);
                exit(0);
        }
    }
    
    sock = getsocket(iname, (size_t*)&nlen, port, NULL, fn);
    
    if ( bind (sock, iname, nlen))
    {
	perror ("bind");
	exit (0);
    }
    puts("Ready");

    dfunc = (fn == NULL) ? display_inet : display_unix;    

    for(done = 0; !done;)
    {
	int n;
	int xlen = olen;
        
        if(listen (sock, 5))
        {
	    perror ("listen");
	    exit(0);
        }

        asock = accept (sock, oname, &xlen);

        printf("sock = %d %d\n", sock, asock);
        
        if (asock == -1)
        {
	    perror ("accept");
	    exit(0);
        }
        xlen = olen;
	getpeername (asock, oname, &xlen);
        (dfunc)(iname, oname);

        while((n = recv (asock, msg, sizeof (msg), 0)) > 0)
        {
            if(strncasecmp (msg, "quit", 4) == 0)
            {
                done = 1;
                break;
            }
	    send(asock, msg, n, 0);
        }
        close(asock);
    }
    close (sock);
    return 0;
}
